/*-----------------------------------------------------------------
LOG
    GEM - Graphics Environment for Multimedia

    A background

    Copyright (c) 1997-1998 Mark Danks. mark@danks.org
    For information on usage and redistribution, and for a DISCLAIMER OF ALL
    WARRANTIES, see the file, "GEM.LICENSE.TERMS" in this distribution.

-----------------------------------------------------------------*/

#ifndef INCLUDE_FILLBACK_H_
#define INCLUDE_FILLBACK_H_

#include "Base/GemShape.h"

/*-----------------------------------------------------------------
-------------------------------------------------------------------
CLASS

    fillback

DESCRIPTION
    
-----------------------------------------------------------------*/
class GEM_EXTERN fillback : public GemShape
{
    CPPEXTERN_HEADER(fillback, GemShape)

    public:

        //////////
        // Constructor
    	fillback(t_floatarg width, t_floatarg height);
    	
    protected:
    	
    	//////////
    	// Destructor
    	virtual ~fillback();

    	//////////
    	// Do the rendering
    	virtual void 	render(GemState *state);

    	//////////
    	// How the object should be drawn
    	virtual void	typeMess(t_symbol *type);

		//////////
		float			m_width;
		float			m_height;
};

#endif	// for header file
