/*-----------------------------------------------------------------
LOG
    GEM - Graphics Environment for Multimedia

    A particle system

    Copyright (c) 1997-1998 Mark Danks. mark@danks.org
    For information on usage and redistribution, and for a DISCLAIMER OF ALL
    WARRANTIES, see the file, "GEM.LICENSE.TERMS" in this distribution.

-----------------------------------------------------------------*/

#ifndef INCLUDE_PARTICLE_H_
#define INCLUDE_PARTICLE_H_

#include "Base/GemBase.h"

/*-----------------------------------------------------------------
-------------------------------------------------------------------
CLASS
	particle
    
	A particle system

DESCRIPTION
    

-----------------------------------------------------------------*/
class GEM_EXTERN particle : public GemBase
{
    CPPEXTERN_HEADER(particle, GemBase)

    public:

	    //////////
	    // Constructor
    	particle();
    	
    	//////////
    	virtual void 	render(GemState *state);

    protected:
    	
    	//////////
    	// Destructor
    	virtual ~particle();

		//////////
		// The particle group
		int				m_particle;
};

#endif	// for header file
