/* ft_conf.h for AmigaOS using ADE enviroment */

/* Define to empty if the keyword does not work. */

/* #undef const */

/* Define if you have a working `mmap' system call. */

/* #undef HAVE_MMAP */

/* Define if you have the ANSI C header files. */

#define STDC_HEADERS 1

/* Define if your processor stores words with the most significant

byte first (like Motorola and SPARC, unlike Intel and VAX). */

#define WORDS_BIGENDIAN 1

/* Define if the X Window System is missing or not being used. */

#define X_DISPLAY_MISSING

/* The number of bytes in a int. */

#define SIZEOF_INT 4

/* The number of bytes in a long. */

#define SIZEOF_LONG 4

/* Define if you have the basename function. */

#define HAVE_BASENAME 1

/* Define if you have the getpagesize function. */

#define HAVE_GETPAGESIZE 1

/* Define if you have the memcpy function. */

#define HAVE_MEMCPY 1

/* Define if you have the valloc function. */

#define HAVE_VALLOC 1

/* Define if you have the <fcntl.h> header file. */

#define HAVE_FCNTL_H 1

/* Define if you have the <unistd.h> header file. */

#define HAVE_UNISTD_H 1
