@echo off

echo Installing GEM...

mkdir ..\doc\gem.examples
mkdir ..\doc\gem.manual
mkdir ..\doc\gem.reference

echo Copying examples...
copy examples\* ..\doc\gem.examples > tempInstall.trash
echo Copying documentation...
copy doc\gem*.* ..\doc\gem.manual > tempInstall.trash
echo Copying help files...
copy help\* ..\doc\gem.reference > tempInstall.trash

del tempInstall.trash

echo Install finished.
