#! /bin/sh
#
# Install GEM
#
# Mark Danks - mark@danks.org
#

echo Installing GEM...

mkdir ../doc/gem.examples
mkdir ../doc/gem.manual
mkdir ../doc/gem.reference

echo Copying examples...
cp examples/* ../doc/gem.examples
echo Copying documentation...
cp doc/* ../doc/gem.manual
echo Copying help files...
cp help/* ../doc/gem.reference


