(*                                                                   *)
(* TTConfig.Inc                                                      *)
(*                                                                   *)
(* This file contains several definition pragmas that are used to    *)
(* build several versions of the library. Each constant is commented *)

{ $DEFINE DEBUG}

(* Define the DEBUG constant if you want the library dumping trace *)
(* information to the standard error output.                       *)

{ $DEFINE ASSERT}

(* Define the ASSERT constant if you want to generate runtime integrity  *)
(* checks within the library. Most of the checks will panic and stop the *)
(* the program when failed..                                             *)

{$IFDEF VIRTUAL}
{$DEFINE INLINE}
{$ENDIF}

(* Define the INLINE constant if you want to use inlining when provided *)
(* by your compiler. Currently, only Virtual Pascal does                *)

{$IFDEF DELPHI}
{$DEFINE USE32}
{$ENDIF}

(* Define the USE32 constant on 32-bit systems. Virtual Pascal *)
(* always define it by default                                 *)

