/*
 * ttflib.h
 *
 * This file is part of the ttf2pk package.
 *
 * Code <-> Postscript tables
 *
 *    Loyer Frederic <loyer@ensta.fr>
 */

void
TTFopen(char *filename, int new_dpi, int new_ptsize, 
        int new_extend, int new_slant);

int
TTFprocess(int Code, byte **bitmap,
           int *width, int *height, int *hoff, int *voff);


/* end */
